//
// N A M E : s n i f f _ c l a s s i c . C
//
// D E S C : simple sniffer using obsolete calls
//
// A U T H O R : Wataru Nishida, MD, PhD
//               wnishida@skyfree.org
//
// M A K E : gcc -Wall -o csniff sniff_classic.C
//
// D A T E : March  3rd, 2000  first revision
//           March 18th, 2001  extensively rewritten
//

#include <sys/types.h>		// socklen_t
#include <sys/socket.h>		// socket(), recvfrom(), SOCK_, PF_
#include <net/ethernet.h>	// ETH_P_
#include <net/if.h>		// IFF_
#include <netinet/in.h>		// htons()

#include <sys/ioctl.h>		// ioctl()
#include <unistd.h>		// close()
#include <stdio.h>		// printf()
#include <string.h>		// strncpy(), strcpy()
#include <signal.h>		// signal(), SIG_

// Compile options

#define STANDALONE		// Do not call external dump_packet()

// Constant definition

const int MaxPacket = 2000;	// Maximum packet size

// GLobal variable

int  SockDes;			// Socket descriptor
char IfcName[ 16 ];		// Interface name holder

// Local definitions

void  usage(void);
void  error_handler(int, const char*);
void  output_packet(const char*, int);
void* cleanup(int);
#ifdef STANDALONE
void  dump_packet(const char*, int);
#else
extern void dump_packet(const char*, int);
#endif

//
// m a i n
//
//   --- Welcome to csniff!
//
//						March 20, 2001

int main(int argc, char* argv[])
 {
  int len,		// Packet length
      ret;		// Return code
  char buf[MaxPacket];	// Packet buffer
  struct ifreq ifr;	// Interface request structure
  struct sockaddr from;	// "socket from" address
  socklen_t fromlen;	// "socket from" length

  // Check arguments

  if (argc != 2)
   { usage(); exit(0); }

  // Set up signal environment.

  signal(SIGHUP,  SIG_IGN);
  signal(SIGINT,  (void(*)(int)) cleanup);
  signal(SIGTERM, (void(*)(int)) cleanup);
  signal(SIGKILL, (void(*)(int)) cleanup);
  signal(SIGQUIT, (void(*)(int)) cleanup);

  // Open a socket with PACKET mode and capture ALL PROTOCOLS.
  // This is OBSOLETE coding style, not recommended!

  SockDes = socket(PF_INET, SOCK_PACKET, htons(ETH_P_ALL));
  printf("Socket was opened as an OBSOLETE PACKET mode.\n");
  error_handler(SockDes, "Can't open a socket!\n");

  // Get current "Interface" flags

  strncpy(IfcName, argv[1], sizeof(IfcName));
  strcpy(ifr.ifr_name, IfcName);
  ret = ioctl(SockDes, SIOCGIFFLAGS, &ifr);
  error_handler(ret, "Can't get interface flags!");

  // Turn on promiscuous mode

  ifr.ifr_flags |= IFF_PROMISC;
  ret = ioctl(SockDes, SIOCSIFFLAGS, &ifr);
  error_handler(ret, "Can't turn on promiscuous mode!");
  printf("%s entered promiscuous mode.\n", IfcName);

  // Receive and display packets forever

  while (1) {
    len = recvfrom(SockDes, buf, MaxPacket, 0, &from, &fromlen);
    error_handler(len, "recvfrom() failed!");

    // NOTICE: Packets are collected from all of interfaces including
    //         loop, eth*, ppp* and so on.
    //
    // "from.sa_data[]" tells us interface name where the packet came from.

    if (strcmp(IfcName, from.sa_data) == 0)	// Filter out packets from
      dump_packet(buf, len);			// other interfaces.
   }
 }

// Turn off promiscuous mode and close the socket

void* cleanup(int sig)
 {
  struct ifreq ifr;	// Interface request structure
  int ret;

  // Get current "Interface" flags

  strcpy(ifr.ifr_name, IfcName);
  ret = ioctl(SockDes, SIOCGIFFLAGS, &ifr);
  if (ret < 0) {
    printf("Can't get interface flags!");
    exit(1);
   }

  // Turn off promiscuous mode

  ifr.ifr_flags &= (~ IFF_PROMISC);
  ret = ioctl(SockDes, SIOCSIFFLAGS, &ifr);
  if (ret < 0) {
    printf("Can't turn off promiscuous mode!");
    exit(1);
   }
  printf("%s left promiscuous mode.\n", IfcName);

  // Close the socket

  close(SockDes);

  exit(0);
 }

// Show usage

void usage(void)
 {
  printf("Usage: csniff interface\n\
            interface; interface name (eth0, eth1..)\n");
 }

// In case of fatal error, report the situation and abort

void error_handler(int result, const char* msg)
 {
  if (result < 0)
   {
    printf("%s\n", msg);
    cleanup(0);
   }
 }

#ifdef STANDALONE
// The simplest packet output routine is here

void dump_packet(const char* buf, int len) {
  printf("*** %d bytes ***\n", len);
  for (int i = 0; i < len; i++)
    printf("%02X ", buf[ i ] & 255);
  printf("\n");
 }
#endif
