//
// N A M E : s n i f f _ c l a s s i c . C
//
// D E S C : simple sniffer using Linux-specific "packet" socket calls
//
// A U T H O R : Wataru Nishida, MD, PhD
//               wnishida@skyfree.org
//
// M A K E : gcc -Wall -o csniff sniff_classic.C
//
// D A T E : March  16th, 2001
//

#include <sys/types.h>		// socklen_t
#include <sys/socket.h>		// socket(), recvfrom(), SOCK_, PF_
#include <net/ethernet.h>	// ETH_P_
#include <net/if.h>		// IFF_
#include <netinet/in.h>		// htons()
#include <netpacket/packet.h>	// packet_mreq, PACKET_

#include <sys/ioctl.h>		// ioctl()
#include <unistd.h>		// close()
#include <stdio.h>		// printf()
#include <string.h>		// strncpy(), strcpy()
#include <signal.h>		// signal(), SIG_

// Compile options

#define STANDALONE		// Do not call external dump_packet()

// Constant definition

const int MaxPacket = 2000;	// Maximum packet size

// GLobal variable

int  SockDes;			// Socket descriptor
char IfcName[ 16] ;		// Interface name holder
int  IfcIdx;			// Current interface's index number
unsigned char HwAddr[ 6 ];	// MAC address holder

// Local definitions

void  usage(void);
void  error_handler(int, const char*);
void  output_packet(const char*, int);
void* cleanup(int);
#ifdef STANDALONE
void  dump_packet(const char*, int);
#else
extern void dump_packet(const char*, int);
#endif

//
// m a i n
//
//   --- Welcome to psniff!
//
//						March 20, 2001

int main(int argc, char* argv[])
 {
  int len,			// Packet length
      ret;			// Return code
  char buf[MaxPacket];		// Packet buffer
  struct ifreq ifr;		// Interface request structure
  struct packet_mreq mreq;      // Membership request structure
  struct sockaddr from;		// "socket from" address
  socklen_t fromlen;		// "socket from" length

  // Check arguments

  if (argc != 2)
   { usage(); exit(0); }

  // Set up signal environment.

  signal(SIGHUP,  SIG_IGN);
  signal(SIGINT,  (void(*)(int)) cleanup);
  signal(SIGTERM, (void(*)(int)) cleanup);
  signal(SIGKILL, (void(*)(int)) cleanup);
  signal(SIGQUIT, (void(*)(int)) cleanup);

  // Open a socket with PF_PACKET mode (It's NEW and Linux-specific !!!)
  // and capture ALL PROTOCOLS.

  SockDes = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
  printf("Socket was opened as a SOCKET RAW mode.\n");
  error_handler(SockDes, "Can't open a socket!\n");

  // Get index number of specified interface

  strncpy(IfcName, argv[1], sizeof(IfcName));
  strcpy(ifr.ifr_name, IfcName); 
  ret = ioctl(SockDes, SIOCGIFINDEX, &ifr);
  error_handler(ret, "Interface doesn't exist!\n");
  IfcIdx = ifr.ifr_ifindex;		// Save index number

  // Get interface hardware address

  ret = ioctl(SockDes, SIOCGIFHWADDR, &ifr);
  error_handler(ret, "Can't get hardware address!");
  for (int i = 0; i < 6; i++)           // Set hardware address
    HwAddr[i] = mreq.mr_address[i] = ifr.ifr_addr.sa_data[i] & 255;

  mreq.mr_ifindex = IfcIdx;		// Set interface index
  mreq.mr_type = PACKET_MR_PROMISC;	// Promiscuous operation
  mreq.mr_alen = 6;			// Set address length

  // Turn on promiscuous mode

  ret = setsockopt(SockDes, SOL_PACKET, PACKET_ADD_MEMBERSHIP, &mreq, sizeof(mreq));
  error_handler(ret, "Failed to activate promiscuous mode\n");
  printf("%s entered promiscuous mode.\n", IfcName);

  // Receive and display packets forever

  while (1) {
    len = recvfrom(SockDes, buf, MaxPacket, 0, &from, &fromlen);
    error_handler(len, "recvfrom() failed!");

    // NOTICE: Packets are collected from all of interfaces including
    //         lo, eth*, ppp* and so on.
    //
    // "from.sa_data[]" tells us interface name where the packet came from.

    if (strcmp(IfcName, from.sa_data) == 0)	// Filter out packets from
      dump_packet(buf, len);			// other interfaces.
   }
 }

// Turn off promiscuous mode and close the socket

void* cleanup(int sig)
 {
  struct packet_mreq mreq;		// Membership request structure
  int ret;

  // Setup multicast request structure

  mreq.mr_ifindex = IfcIdx;		// Set interface index
  mreq.mr_type = PACKET_MR_PROMISC;	// Promiscuous operation
  mreq.mr_alen = 6;			// Set address length
  for (int i = 0; i < 6; i++)		// Set hardware address
    mreq.mr_address[ i ] = HwAddr[ i ];

  // Turn off promiscuous mode

  ret = setsockopt(SockDes, SOL_PACKET, PACKET_DROP_MEMBERSHIP, &mreq, sizeof(mreq));
  error_handler(ret, "Failed to activate promiscuous mode\n");
  printf("%s left promiscuous mode.\n", IfcName);
  
  // Close the socket

  close(SockDes);

  exit(0);
 }

// Show usage

void usage(void)
 {
  printf("Usage: psniff interface\n\
            interface; interface name (eth0, eth1..)\n");
 }

// In case of fatal error, report the situation and abort

void error_handler(int result, const char* msg)
 {
  if (result < 0)
   {
    printf("%s\n", msg);
    cleanup(0);
   }
 }

#ifdef STANDALONE
// The simplest packet output routine is here

void dump_packet(const char* buf, int len) {
  printf("*** %d bytes ***\n", len);
  for (int i = 0; i < len; i++)
    printf("%02X ", buf[ i ] & 255);
  printf("\n");
 }
#endif
